/* 
 * scannedonlyd scanserver
 *
 * Copyright 2007,2008,2009,2010 (C) Olivier Sessink
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*#define DEBUG*/
/*#define CHECK_CONSISTENCY*/

#define _POSIX_C_SOURCE 200112L

#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <regex.h>

#include <pthread.h>
#include <clamav.h>
#include "config.h"
#ifdef DEBUG
#define DEBUG_MSG printf
#else
#define DEBUG_MSG(args...)		/* */
#endif

#define BUFFERSIZE 1500 /* default MTU */
#ifndef SUN_LEN
#define SUN_LEN(su) (sizeof (*(su)) - sizeof ((su)->sun_path) + strlen((su)->sun_path))
#endif
#define INT_TO_POINTER(i) ((void *) (long) (i))
#define POINTER_TO_INT(p) ((int)  (long) (p))

typedef struct {
	char *filename;
	void *next;

} Tlist;


typedef struct {
	Tlist * first;
	Tlist * last;
	int len;

} Tqueue;

pthread_spinlock_t statuslock;

pthread_mutex_t queuelock;
pthread_cond_t queuechanged;

Tqueue * small_queue;

Tqueue * large_queue;
char **currently_scanning;

int large_scanning_threads = 0;

int maxthreads = 4, maxlargethreads = 1, bigfiles = 10;
char *quarantainedir = NULL;

struct cl_engine *engine[2] = { NULL, NULL };
int engine_users[2] = { 0, 0 };

int current_engine = 0;
pthread_mutex_t enginelock;
struct timespec next_dbcheck;
int dbcheck_timewait_thread = 0;
#ifndef HAVE_LIBCLAMAV_CL_INIT
struct cl_limits limits;
#endif
struct cl_stat dbstat;
int dbcheck_timer = 5;

int queue_limit = 1000;

struct timespec prev_status_update, next_status_update;
int status_timewait_thread = 0;
int status_num_files = 0, status_num_files_large = 0, status_num_skipped =
	0, status_num_skipped_large = 0, status_num_failed =
	0, status_outside_scanroot = 0, status_max_small_queue =
	0, status_max_large_queue = 0, status_min_small_queue = 0, status_min_large_queue = 0;
int status_timer = 10;			/* minutes */

static void push_to_queue(Tqueue * q, char *filename)
{
	Tlist *l;
	if (!q || !filename) {
		DEBUG_MSG("\n\n**** WARNING, q=%p, filename=%p\n\n\n",q,filename);
#ifdef DEBUG
	exit(666);
#endif
		return;
	}
	
	l = malloc(sizeof(Tlist));
	l->filename = strdup(filename);
	l->next = NULL;
	if (q->last)
		q->last->next = l;
	if (q->first == NULL)
		q->first = l;
	q->last = l;
	q->len++;
#ifdef CHECK_CONSISTENCY
		{
			Tlist * l2;
			int count = 0;
			l2 = q->first;
			while (l2) {
				count++;
				l2 = l2->next;
			}
			printf("push_to_queue() count=%d, len=%d\n", count, q->len);
#ifdef DEBUG
			if (count != q->len) {
				printf("\n\n\n********************* QUEUE ERROR ******************\n\n\n");
				exit(666);
			}
#endif
		}		
#endif	/*  */

}

static char *pop_from_queue(Tqueue * q)
{
	if (q->first) {
		char *filename;
		Tlist * l = q->first;
		if (q->first == q->last)
			q->last = NULL;
		if (q->first)
			q->first = q->first->next;
		q->len--;
		
#ifdef CHECK_CONSISTENCY
		{
			Tlist * l2;
			int count = 0;
			l2 = q->first;
			while (l2) {
				count++;
				l2 = l2->next;
			}
			printf("pop_from_queue() count=%d, len=%d\n", count, q->len);
#ifdef DEBUG
			if (count != q->len) {
				printf("\n\n\n********************* QUEUE ERROR ******************\n\n\n");
				exit(666);
			}
#endif
		}		
#endif	/*  */
		filename = l->filename;
		free(l);
		return filename;
	}
	printf("warning, tried to pop from empty queue!!!!!!!!\n");
#ifdef DEBUG
	exit(666);
#endif
	return NULL;
}

static void remove_filename_from_queue(Tqueue *q, const char *filename) {
	Tlist *prev=NULL,*cur,*next;
	
	if (!q || !filename)
		return;
	
	cur = q->first;
	while (cur) {
		next=cur->next;
		if (strcmp(filename, cur->filename)==0) {
			if (prev)
				prev->next=next;
			if (cur == q->first)
				q->first = next;
			if (cur == q->last)
				q->last = prev;
			q->len--;
			DEBUG_MSG("removed entry for filename %s from queue\n",filename);
			free(cur->filename);
			free(cur);
		} else {
			prev=cur;
		}
		cur = next;
	}
}

static Tqueue *new_queue(void)
{
	Tqueue * q = malloc(sizeof(Tqueue));
	q->first = NULL;
	q->last = NULL;
	q->len = 0;
	return q;
}

static int rename_or_move(const char *from, const char *to)
{
	int fromfd, tofd, bytes;
	unsigned int size = 0;
	char buf[4096];
	struct stat sbuf;

	if (rename(from, to) == 0)
		return 0;

	if (errno != EXDEV) {
		syslog(LOG_NOTICE, "failed to rename %s to %s %d: %s\n", from, to, errno, strerror(errno));
		return -1;
	}

	/* need copy && unlink */
	errno = 0;
	if (stat(from, &sbuf) < 0) {
		syslog(LOG_NOTICE,
			   "when trying to copy %s to %s, failed to stat %s %d: %s\n",
			   from, to, from, errno, strerror(errno));
		return -1;
	}

	fromfd = open(from, O_RDONLY);
	if (!fromfd) {
		syslog(LOG_NOTICE,
			   "when trying to copy %s to %s, failed to open %s %d: %s\n",
			   from, to, from, errno, strerror(errno));
		return -1;
	}
	tofd = open(to, O_WRONLY | O_CREAT, 00660);
	if (!tofd) {
		syslog(LOG_NOTICE,
			   "when trying to copy %s to %s, failed to open %s %d: %s\n",
			   from, to, to, errno, strerror(errno));
		return -1;
	}
	errno = 0;
	while ((bytes = read(fromfd, buf, sizeof(buf))) > 0) {
		write(tofd, buf, bytes);
		size += bytes;
	}
	close(fromfd);
	close(tofd);
	if (sbuf.st_size != size) {
		syslog(LOG_NOTICE,
			   "when trying to copy %s to %s, wrote only %d of %d bytes\n",
			   from, to, (unsigned int) size, (unsigned int) sbuf.st_size);
		unlink(to);
		return -1;
	}
	if (unlink(from) < 0) {
		syslog(LOG_NOTICE,
			   "when trying to copy %s to %s, failed to unlink %s %d: %s\n",
			   from, to, from, errno, strerror(errno));
		return -1;
	}
	return 0;
}

static char *normalize_filename(char *filename)
{
	int src = 0, dst = 0;
	if (!filename)
		return NULL;

	/* loop over the filename */
	while (filename[src] != '\0') {
		/*printf("src=%d, dst=%d, filename[%d]=%c, filename[%d]=%c\n",src,dst,src,filename[src],dst,filename[dst]); */
		if (filename[src] == '/' && filename[src + 1] == '.' && filename[src + 2] == '/') {
			src += 2;
			continue;
		}
		if (filename[src] == '/' && filename[src + 1] == '/') {
			src++;
			continue;
		}
		if (src != dst) {
			filename[dst] = filename[src];
		}
		dst++;
		src++;
	}
	filename[dst] = '\0';
	return filename;
}


static char *filename_backend(const char *fullpath, const char *prefix, const char *suffix)
{
	char *tmp, *cachefile, *shortname;
	int rlen, blen, plen, slen, totlen;
	tmp = strrchr(fullpath, '/');
	if (!tmp)
		return NULL;
	blen = (tmp - fullpath) + 1;
	shortname = tmp + 1;
	rlen = strlen(shortname);
	slen = strlen(suffix);
	plen = strlen(prefix);
	totlen = rlen + blen + plen + slen + 1;
	cachefile = (char *) malloc(totlen);
	memset(cachefile, 0, totlen);
	cachefile =
		strncat(strncat
				(strncat
				 (strncpy(cachefile, fullpath, blen), prefix, plen + 1),
				 shortname, rlen + 1), suffix, slen + 1);
	return cachefile;
}

static char *cachefile_name(const char *fullpath)
{
	return filename_backend(fullpath, ".scanned:", "");
}

static void create_file(const char *name, mode_t mode, uid_t uid, gid_t gid)
{
	int fd = open(name, O_CREAT | O_WRONLY | O_EXCL, mode);
	if (fd == -1) {
		if (errno == EEXIST) {
			struct timeval ts[2];
			gettimeofday(&ts[0], NULL);
			
				/* touch the file */ 
				/*ts[0].tv_usec = 0;
				   ts[0].tv_sec = time(NULL); */ 
				ts[1] = ts[0];
			utimes(name, ts);
		} else {
			syslog(LOG_WARNING, "Failed to create %s", name);
		}
	} else if (fd > 0) {
		close(fd);
		chown(name, uid, gid);
	}
}

/* this function is always called within the queuelock, all values are only updated in the queuelock 
except for some values that change in the statuslock. That's why we ALSO aquire the statuslock.
There is no other code that aquires BOTH the queuelock AND the statuslock so this cannot lead to
a deadlock situation */
static void do_status_update(int id)
{
	long int timespan;
	int i;
	timespan = (next_status_update.tv_sec - prev_status_update.tv_sec) / 60;
	DEBUG_MSG("do_status_update()\n");
	syslog(LOG_NOTICE,
		   "Status: scanned %d files and %d large files in %ld minutes; min %d/max %d requests on small queue, min %d/max %d requests on large queue",
		   status_num_files, status_num_files_large, timespan,
		   status_min_small_queue, status_max_small_queue, status_min_large_queue, status_max_large_queue);
	if (status_num_skipped || status_num_skipped_large) {
		syslog(LOG_WARNING,
			   "Warning: %d requests and %d large file requests were skipped in %ld minutes (too busy)",
			   status_num_skipped, status_num_skipped_large, timespan);
	}
	if (status_num_failed) {
		syslog(LOG_WARNING,
			   "Warning: %d files failed to scan in %ld minutes (error or virus found)",
			   status_num_failed, timespan);
	}
	if (status_outside_scanroot) {
		syslog(LOG_NOTICE,
			   "Note: %d requests were not in the scanroot in %ld minutes",
			   status_outside_scanroot, timespan);
	}
	for (i = 0; i < maxthreads; i++) {
		if (i == id)
			syslog(LOG_DEBUG, "Info: thread %d is logging this status update", i);
		else {
			if (currently_scanning[i])
				syslog(LOG_DEBUG, "Info: thread %d is currently scanning %s", i, currently_scanning[i]);
			else
				syslog(LOG_DEBUG, "Info: thread %d is idle", i);
		}
	}

	pthread_spin_lock(&statuslock);
	status_num_files = 0;		/* accessed by all scanning threads after finishing scanning and status update */
	status_num_files_large = 0;	/* accessed by all scanning threads after finishing scanning and status update */
	status_num_failed = 0;		/* accessed by all scanning threads and status update */
	status_num_skipped = 0;		/* accessed by main thread and status update */
	status_num_skipped_large = 0;	/* accessed by all scanning threads if pushed to large queue and status update */
	status_min_small_queue = status_max_small_queue = small_queue->len;
	status_min_large_queue = status_max_large_queue = large_queue->len;
	status_outside_scanroot = 0;
	pthread_spin_unlock(&statuslock);

	prev_status_update.tv_sec = time(NULL);
	next_status_update.tv_sec = prev_status_update.tv_sec + 60 * status_timer;	/* in X minutes we give the next status message */
	next_status_update.tv_nsec = 0;

}

static void handle_file_warning(int logid, const char *file,
								const char *warnprefix, int do_move,
								const char *moveprefix, mode_t mode, uid_t uid, gid_t gid)
{
	char *warn, *movename;
	if (!file || file[0] != '/') {
		/* defensive programming */
		syslog(LOG_CRIT, "internal error in handle_file_warning. This is a BUG, notify the developers\n");
		return;
	}
	warn = filename_backend(file, warnprefix, ".txt");
	create_file(warn, mode, uid, gid);
	if (!do_move) {
		free(warn);
		return;
	}
	if (quarantainedir) {
		int qlen, flen, count = 0, cont = 1;
		char *shortname;
		shortname = strrchr(file, '/');
		shortname++;

		qlen = strlen(quarantainedir);
		flen = strlen(shortname);
		movename = malloc(qlen + flen + 10);
		/* 10 means we can append 999999999 */
		memset(movename, 0, (qlen + flen + 10));
		movename = strcat(strncpy(movename, quarantainedir, qlen), shortname);
		while (cont) {
			struct stat sbuf;
			if (stat(movename, &sbuf) == 0) {
				snprintf(movename, qlen + flen + 10, "%s%s%d", quarantainedir, shortname, count);
				count++;
			} else {
				cont = 0;
			}
		}
	} else {
		movename = filename_backend(file, moveprefix, "");
	}

	syslog(LOG_NOTICE, "%d: moving %s to %s", logid, file, movename);
	if (rename_or_move(file, movename) != 0) {
		syslog(LOG_WARNING, "failed to rename %s to %s, error %d: %s",
			   file, movename, errno, strerror(errno));
	}
	free(movename);
	free(warn);
}

static int is_currently_scanning(const char *filename, int my_id)
{
	int i;

	if (!filename)
		return 0;

	for (i = 0; i < maxthreads; i++) {
		if (i != my_id && currently_scanning[i]
			&& strcmp(filename, currently_scanning[i]) == 0) {
			return 1;
		}
	}
	return 0;
}

static struct cl_engine *load_engine()
{
	int ret;
	unsigned int sigs = 0;
	struct cl_engine *tmp = NULL;
	DEBUG_MSG("load_engine()\n");
#ifdef HAVE_LIBCLAMAV_CL_INIT
	tmp = cl_engine_new();
	if (!tmp) {
		syslog(LOG_CRIT, "WARNING: Failed to create new ClamAV engine");
	}
	cl_engine_set_num(tmp, CL_ENGINE_MAX_SCANSIZE, (long long) 1024 * 1024 * 1024 * 5);
	cl_engine_set_num(tmp, CL_ENGINE_MAX_FILESIZE, (long long) 1024 * 1024 * 1024 * 5);
	if ((ret = cl_load(cl_retdbdir(), tmp, &sigs, CL_DB_STDOPT)) != CL_SUCCESS) {
		syslog(LOG_CRIT, "WARNING: Loading ClamAV database failed: %s", cl_strerror(ret));
		cl_engine_free(tmp);
		return NULL;
	}
	if ((ret = cl_engine_compile(tmp)) != CL_SUCCESS) {
		syslog(LOG_CRIT, "WARNING: Building ClamAV engine failed: %s", cl_strerror(ret));
		cl_engine_free(tmp);
		return NULL;
	}
#else
	if ((ret = cl_load(cl_retdbdir(), &tmp, &sigs, CL_DB_STDOPT))) {
		syslog(LOG_CRIT, "WARNING: Loading ClamAV database failed: %s", cl_strerror(ret));
		return NULL;
	}
	if ((ret = cl_build(tmp))) {
		syslog(LOG_CRIT, "WARNING: Building ClamAV engine failed: %s", cl_strerror(ret));
		cl_free(tmp);
		return NULL;
	}
#endif
	return tmp;
}

static void do_dbcheck(int id)
{
	int need_new;
	int newnum, oldnum = current_engine;
	DEBUG_MSG("do_dbcheck()\n");
	newnum = (current_engine == 1) ? 0 : 1;
	if (engine_users[newnum] != 0) {
		/* try next minute */
		next_dbcheck.tv_sec = time(NULL) + 60 * 1;
		return;
	}
	/* check if we need a new engine */
	need_new = cl_statchkdir(&dbstat);
	/* schedule next check in dbcheck_timer minutes */
	next_dbcheck.tv_sec = time(NULL) + 60 * dbcheck_timer;
	if (!need_new) {
		return;
	}
	syslog(LOG_INFO, "%d: New ClamAV database available, starting reload...", id);
	/* free the old engine, SHOULD HAVE BEEN DONE ALREADY */
	if (engine[newnum]) {
		syslog(LOG_CRIT,
			   "BUG, engine_users[%d]=%d, so engine[%d] should not be %p\n",
			   newnum, engine_users[newnum], newnum, engine[newnum]);
#ifdef HAVE_LIBCLAMAV_CL_INIT
		cl_engine_free(engine[newnum]);
#else
		cl_free(engine[newnum]);
#endif
		engine[newnum] = NULL;
	}

	/* load the new engine */
	engine[newnum] = load_engine();
	if (engine[newnum] != NULL) {
		int free_old;
		syslog(LOG_NOTICE,
			   "%d: New ClamAV database loaded; switching from ClamAV engine %d to %d",
			   id, current_engine, newnum);
		pthread_mutex_lock(&enginelock);
		current_engine = newnum;
		free_old = (engine_users[oldnum] == 0);
		pthread_mutex_unlock(&enginelock);
		if (free_old) {
			syslog(LOG_INFO, "%d: ClamAV engine %d is not used, freeing...", id, oldnum);
#ifdef HAVE_LIBCLAMAV_CL_INIT
			cl_engine_free(engine[oldnum]);
#else
			cl_free(engine[oldnum]);
#endif
			engine[oldnum] = NULL;
		}
		cl_statfree(&dbstat);
		cl_statinidir(cl_retdbdir(), &dbstat);
	}
}

static void scanningthread(void *idp)
{
	int scanning_large = 0;
	int id = POINTER_TO_INT(idp);
	for (;;) {
		int ret;
		struct stat sbuf1;
		char *scannedname, *failedname;
		struct stat sbuf2;

		pthread_mutex_lock(&queuelock);	/* <--------------------------------------- LOCK START */
		DEBUG_MSG("%d: have lock, currently_scanning[%i]=%p\n", id, id, currently_scanning[id]);
		if (currently_scanning[id]) {
			DEBUG_MSG("%d: have lock, remove %s from small_queue\n", id, currently_scanning[id]);
			remove_filename_from_queue(small_queue, currently_scanning[id]);
			DEBUG_MSG("%d: have lock, free currently_scanning[%i]=%s\n", id, id, currently_scanning[id]);
			free(currently_scanning[id]);
			currently_scanning[id] = NULL;
		}
		
		/* if our previous scan was a large file scan, reduce the number now */ 
		if (scanning_large) {
			DEBUG_MSG("%d: decreasing scanning_large\n", id);
			large_scanning_threads--;
			scanning_large = 0;
		}
		
			/* now wait for something on the queue */ 
		while (small_queue->len == 0
				   && (large_queue->len == 0 || large_scanning_threads >= maxlargethreads)) {
			/*printf("%d: sleeping, small_queue=%d, large_queue=%d (with %d threads)\n",id,small_queue->len, large_queue->len, large_scanning_threads); */
			if (status_timewait_thread == 0) {
				status_timewait_thread = 1;
				ret = pthread_cond_timedwait(&queuechanged, &queuelock, &next_status_update);
				if (ret == ETIMEDOUT) {
					do_status_update(id);
				}
				status_timewait_thread = 0;
			} else if (dbcheck_timewait_thread == 0) {
				dbcheck_timewait_thread = 1;
				ret = pthread_cond_timedwait(&queuechanged, &queuelock, &next_dbcheck);
				if (ret == ETIMEDOUT) {
					/* unlock the queue first !?!?! */
					pthread_mutex_unlock(&queuelock);
					do_dbcheck(id);
					pthread_mutex_lock(&queuelock);
				}
				dbcheck_timewait_thread = 0;
			} else {
				pthread_cond_wait(&queuechanged, &queuelock);
			}
			
			/*printf("%d: waking up, small_queue=%d, large_queue=%d (with %d threads)\n",id,small_queue->len, large_queue->len, large_scanning_threads); */ 
		}
		if ((status_timewait_thread == 0)
			&& (time(NULL) >= next_status_update.tv_sec)) {
			status_timewait_thread = 1;
			do_status_update(id);
			status_timewait_thread = 0;
		}
		if ((dbcheck_timewait_thread == 0)
			&& (time(NULL) >= next_dbcheck.tv_sec)) {
			dbcheck_timewait_thread = 1;
			pthread_mutex_unlock(&queuelock);
			do_dbcheck(id);
			pthread_mutex_lock(&queuelock);
			dbcheck_timewait_thread = 0;
		}
		
		/* check if we should do a large file scan */ 
		if (large_scanning_threads < maxlargethreads && large_queue->len != 0) {
			currently_scanning[id] = pop_from_queue(large_queue);
			if (status_min_large_queue > large_queue->len)
				status_min_large_queue = large_queue->len;
			DEBUG_MSG("%d: popping %s from large queue\n", id, currently_scanning[id]);
			scanning_large = 1;
			large_scanning_threads++;
			syslog(LOG_DEBUG,
				   "%d: popped %s from large queue, queue length %d", id,
				   currently_scanning[id], large_queue->len);
				/*printf("%d: popped %s from large queue\n",id,name); */ 
		} else {
			/* use regular queue */ 
			currently_scanning[id] = pop_from_queue(small_queue);
			if (status_min_small_queue > small_queue->len)
				status_min_small_queue = small_queue->len;
			DEBUG_MSG("%d: popping %s from small queue\n", id, currently_scanning[id]);
			syslog(LOG_DEBUG,
				   "%d: popped %s from small queue, queue length %d", id,
				   currently_scanning[id], small_queue->len);
				/*printf("%d: popped %s from small queue\n",id,name); */ 
		}

		if (is_currently_scanning(currently_scanning[id], id)) {
			/*printf("%d: file %s is being scanned already: skip!\n",name); */
			DEBUG_MSG("%d: file %s is being scanned already!\n", id, currently_scanning[id]);
			free(currently_scanning[id]);
			currently_scanning[id] = NULL;
		}
		pthread_mutex_unlock(&queuelock);	/* <--------------------------------------- LOCK END */
		if (currently_scanning[id] == NULL) {
			continue;
		}
		
		/* test if the file to scan exists */ 
		ret = stat(currently_scanning[id], &sbuf1);
		if (ret == -1) {
			/* failed to stat this file, how is this possible ?? perhaps it doesn't exist */
			if (errno == ENOENT) {
				syslog(LOG_INFO, "%d: Not scanning %s: it doesn't exist", id, currently_scanning[id]);
			} else {
				syslog(LOG_WARNING,
					   "%d: Error during stat(%s): %s, not scanning", id,
					   currently_scanning[id], strerror(errno));
			}
			continue;
		} 
		if (!S_ISREG(sbuf1.st_mode)) {
			syslog(LOG_INFO, "%d: Not scanning %s: not a regular file", id, currently_scanning[id]);
			continue;
		}
		if (sbuf1.st_size == 0) {
			syslog(LOG_INFO, "%d: Not scanning %s: zero byte file", id, currently_scanning[id]);
			continue;
		}
		
		if (scanning_large == 0 && sbuf1.st_size > (bigfiles * 1048576) /* Mb */ ) {
			pthread_mutex_lock(&queuelock);	/* <--------------------------------------- LOCK START */
			if (large_scanning_threads >= maxlargethreads) {
				if (large_queue->len < queue_limit) {
					
					/*printf("%d: file %s pushed to large queue\n",id,name); */ 
					push_to_queue(large_queue, currently_scanning[id]);
					if (status_max_large_queue < large_queue->len)
						status_max_large_queue = large_queue->len;
					pthread_mutex_unlock(&queuelock);
					pthread_cond_signal(&queuechanged);
				} else {
					syslog(LOG_INFO,
						   "%d: large queue >= %d: skipping request to scan %s",
						   id, queue_limit, currently_scanning[id]);
					status_num_skipped_large++;
					pthread_mutex_unlock(&queuelock);
				}
				continue;
			} else {
				/*printf("%d: file %s is large, but there is no large file scanned right now\n",id,name); */ 
				large_scanning_threads++;
				scanning_large = 1;
			}
			pthread_mutex_unlock(&queuelock);	/* <--------------------------------------- LOCK END */
		}
		scannedname = cachefile_name(currently_scanning[id]);
		failedname = filename_backend(currently_scanning[id], ".failed:", "");
		ret = stat(scannedname, &sbuf2);
		if (ret == 0) {
			if (!S_ISREG(sbuf2.st_mode)) {
				/* anomaly, unlink it */
				unlink(scannedname);
			}
		} else {			/* if it does not exist, perhaps the previous scan failed, check for that */
			int ret3;
			struct stat sbuf3;

			ret3 = stat(failedname, &sbuf3);
			if (ret3 == 0) {	/* there has been a previous scan that failed */
				if (!S_ISREG(sbuf3.st_mode)) {
					/* anomaly, unlink it */
					unlink(failedname);
				} else {
					/* it is a regular file and it exists, check the timestamps */
					if (sbuf1.st_ctime < sbuf3.st_ctime) {
						/* .failed: file is newest, rescanning is useless */
						if ((sbuf3.st_ctime + 60 * 5) < time(NULL)) {
							/* the file did not change the last 5 minutes, it's probably not actively written to
							   so warn the user  */
							syslog(LOG_INFO,
								   "%d: Not scanning %s, last scan failed >5 minutes ago, creating a warning file",
								   id, currently_scanning[id]);
							handle_file_warning(id,currently_scanning[id],"failed_to_virus-scan_",0,
								NULL, sbuf1.st_mode, sbuf1.st_uid, sbuf1.st_gid);
						} else {
							syslog(LOG_INFO,
								   "%d: Not scanning %s, previous scan failed",
								   id, currently_scanning[id]);
						}
						free(scannedname);
						free(failedname);
						continue;
					} else {
						syslog(LOG_INFO,
							   "%d: Will rescan %s, file changed after last failed attempt",
							   id, currently_scanning[id]);
						unlink(failedname);
					}
				}
			}
		}
		if (ret == -1 || (ret == 0 && (sbuf1.st_ctime > sbuf2.st_ctime))) {
			/* either the .scanned: file does not exist, or the file is newer than the .scanned: file
			   in any way we need to re-scan it */ 
			const char *virname = NULL;
			int enginenum;
			pthread_mutex_lock(&enginelock);
			enginenum = current_engine;
			engine_users[enginenum]++;
			pthread_mutex_unlock(&enginelock);
			syslog(LOG_INFO, "%d:%d: Scanning %s", id, enginenum, currently_scanning[id]);
#ifdef HAVE_LIBCLAMAV_CL_INIT
			ret = cl_scanfile(currently_scanning[id], &virname, NULL, engine[enginenum], CL_SCAN_STDOPT);
#else
			ret = cl_scanfile(currently_scanning[id], &virname, NULL,
							engine[enginenum], &limits, CL_SCAN_STDOPT);
#endif
			if (ret == CL_VIRUS) {
				status_num_failed++;
				syslog(LOG_WARNING,
					   "%d:%d: WARNING: %s contains virus %s!", id,
					   enginenum, currently_scanning[id], virname);
				handle_file_warning(id, currently_scanning[id],
									"VIRUS_found_in_",1, ".virus:",
									sbuf1.st_mode, sbuf1.st_uid, sbuf1.st_gid);
				unlink(scannedname);

			} else if (ret == CL_CLEAN) {
				/*printf("%d: %s is clean, creating %s\n",id,name,scannedname); */
				syslog(LOG_INFO, "%d:%d: %s is clean", id, enginenum, currently_scanning[id]);
				create_file(scannedname, sbuf1.st_mode, sbuf1.st_uid, sbuf1.st_gid);
			} else {		/* if(ret != CL_CLEAN) */
				status_num_failed++;
				syslog(LOG_CRIT,
					   "%d:%d: WARNING: failed to scan %s: %s", id,
					   enginenum, currently_scanning[id], cl_strerror(ret));
				create_file(failedname, sbuf1.st_mode, sbuf1.st_uid, sbuf1.st_gid);
				unlink(scannedname);
			}
			pthread_mutex_lock(&enginelock);
			engine_users[enginenum]--;
			if (enginenum != current_engine && engine_users[enginenum] == 0) {
				syslog(LOG_INFO, "%d: Last user of ClamAV engine %d, freeing...", id, enginenum);
#ifdef HAVE_LIBCLAMAV_CL_INIT
				cl_engine_free(engine[enginenum]);
#else
				cl_free(engine[enginenum]);
#endif
				engine[enginenum] = NULL;
			}
			pthread_mutex_unlock(&enginelock);

			pthread_spin_lock(&statuslock);
			if (scanning_large)
				status_num_files_large++;
			else
				status_num_files++;
			pthread_spin_unlock(&statuslock);
		}
		
		free(scannedname);
		free(failedname);
		/*printf("%d: done with %s, next loop...\n",id,name); */ 
	}
}

static void print_usage(void)
{
	printf("\nUsage: scannedonlyd_clamav <options>\n\n");
	printf("--portnum -p   : specify the UDP port to listen on (default 2020)\n");
	printf("--socket -s    : specify socket name (default /var/lib/scannedonly/scan)\n");
	printf("--maxthreads -t: specify total number of scanning threads (default 4)\n");
	printf("--maxlargethreads -g: specify max. number of scanning threads large files (default 1)\n");
	printf("--loglevel -l  : specify log level 0-3 (default 1)\n");
	printf("--big -b       : specify which size (Mb) files are treated as big (default 5)\n");
	printf("--time -m      : specify the interval (minutes) for status messages (default 10)\n");
	printf("--scanroot -r  : specify the root of the directories to scan (default /)\n");
	printf("--quarantainedir -d: specify a directory where to move infected and failed files (default no move, prepend prefix)\n");
	printf("--exclude -e   : specify a (posix) regular expression pattern to exclude certain files (default no excludes)\n");
	printf("--queuelen -q  : specify the maximum queue length (default 1000)\n");
	printf("--pidfile -i   : specify the location for a PID file (default none)\n");
	printf("--nodaemon -n  : don't detach from the terminal\n");
	printf("\nuse either -s OR -p but not both. If nothing is specified scannedonlyd will\n");
	printf("listen on the domain socket\n\n");
}

static void write_pidfile(char *pidfile, int pid)
{
	if (pidfile) {
		FILE *fd = fopen(pidfile, "w");
		if (fd) {
			fprintf(fd, "%d", (int) pid);
			fclose(fd);
		} else {
			printf("ERROR: failed to write PID file %s: %s\n", pidfile, strerror(errno));
		}
	}
}

static void detach(char *pidfile)
{
	pid_t pid, sid;
	/* daemonize */
	pid = fork();
	if (pid < 0) {
		printf("ABORT, failed to detach daemon\n");
		exit(5);
	}
	if (pid > 0) {
		/* write the PID file */
		write_pidfile(pidfile, pid);
		/* the parent process will now exit, the child is the real daemon */
		exit(0);
	}
	/* At this point we are executing as the child process */
	sid = setsid();
	chdir("/");
	freopen("/dev/null", "r", stdin);
	freopen("/dev/null", "w", stdout);
	freopen("/dev/null", "w", stderr);

}

static void sighandler(int signal)
{
	if (signal == 11) {
		syslog(LOG_CRIT, "ABORT: segmentation fault");
		exit(1);
	} else {
		syslog(LOG_NOTICE, "Exiting, got signal %d", signal);
		exit(0);
	}
}

struct handle_file_options {
	char *excludepattern;
	char *scanroot;
	unsigned int queue_limit;
	unsigned int scanrootlen;
	regex_t reg_pat;
};

static void handle_filename(struct handle_file_options *hfo, char *filename)
{
	int wakeupcall = 0;
	if (filename == NULL)
		return;
	if (filename[0] != '/') {
		syslog(LOG_INFO, "Ignored request %s; it is not an absolute filename", filename);
		return;
	}
	if (strstr(filename, "/../") != NULL) {
		syslog(LOG_INFO, "Ignored request %s; it contains upward directory references /../ ", filename);
		return;
	}
	filename = normalize_filename(filename);

	if ((hfo->scanroot)
		&& (strncmp(hfo->scanroot, filename, hfo->scanrootlen) != 0)) {
		syslog(LOG_NOTICE,
			   "Ignored request to scan %s; it is outside the scanroot %s", filename, hfo->scanroot);
		status_outside_scanroot++;
		return;
	}
	if (hfo->excludepattern) {
		int retval;
		retval = regexec(&hfo->reg_pat, filename, 0, NULL, 0);
		if (retval == 0) {
			syslog(LOG_INFO,
				   "Ignored request to scan %s; it matches the excludepattern %s",
				   filename, hfo->excludepattern);
			return;
		} else if (retval != REG_NOMATCH) {
			char errbuf[1024];
			errbuf[0] = '\0';
			regerror(retval, &hfo->reg_pat, errbuf, 1024);
			syslog(LOG_CRIT,
				   "Failed to match excludepattern %s with filename %s: %s",
				   hfo->excludepattern, filename, errbuf);
		}
	}
	wakeupcall = 0;
	pthread_mutex_lock(&queuelock);
	if (small_queue->len < hfo->queue_limit) {
		push_to_queue(small_queue, filename);
		syslog(LOG_INFO, "Received %s for scanning, %d files on small queue", filename, small_queue->len);
		if (status_max_small_queue < small_queue->len)
			status_max_small_queue = small_queue->len;
		if (small_queue->len >= maxthreads) {
			/*printf("%d small files on queue, waking up all threads!\n",small_queue->len); */ 
			wakeupcall = 2;
		} else {
			/*printf("waking up a single thread\n"); */ 
			wakeupcall = 1;
		}
	} else {
		syslog(LOG_INFO,
			   "Warning: Received %s but skipped request; small queue limit %d", filename, small_queue->len);
		status_num_skipped++;
	}
	pthread_mutex_unlock(&queuelock);

	if (wakeupcall == 1)
		pthread_cond_signal(&queuechanged);
	else if (wakeupcall == 2)
		pthread_cond_broadcast(&queuechanged);
}


int main(int argc, char **argv)
{
	char buffer[BUFFERSIZE];
	register int i, s, len;
	struct handle_file_options hfo;
	
	pthread_t thread;
	pthread_attr_t attr;
	int domain_socket = 1, portnum = 2020, detach_from_terminal = 1;
	char *socketname = NULL, *pidfile = NULL;
	int c = 0, loglevel = 1;

	hfo.scanroot = NULL;
	hfo.queue_limit = 100000000;
	hfo.excludepattern = NULL;
	while (c != -1) {
		int option_index;
		static struct option long_options[] = {
			{"portnum", required_argument, NULL, 'p'},
			{"socket", required_argument, NULL, 's'},
			{"maxthreads", required_argument, NULL, 't'},
			{"maxlargethreads", required_argument, NULL, 'g'},
			{"loglevel", required_argument, NULL, 'l'},
			{"big", required_argument, NULL, 'b'},
			{"time", required_argument, NULL, 'm'},
			{"scanroot", required_argument, NULL, 'r'},
			{"queuelen", required_argument, NULL, 'q'},
			{"pidfile", required_argument, NULL, 'i'},
			{"exclude", required_argument, NULL, 'e'},
			{"quarantainedir", required_argument, NULL, 'd'},
			{"nodaemon", no_argument, NULL, 'n'},
			{"help", no_argument, NULL, 'h'},
			{NULL, 0, NULL, 0} 
		};
		
		c = getopt_long(argc, argv, "p:s:t:g:l:b:m:r:q:i:e:d:nh?", long_options, &option_index);
		switch (c) {
		case 'p':
			if (socketname != NULL) {
				printf("cannot set domain socket AND UDP port number\n");
				exit(1);
			}
			portnum = atoi(optarg);
			domain_socket = 0;
			break;
		case 's':
			if (domain_socket == 0) {
				printf("cannot set UDP port number AND domain socket\n");
				exit(1);
			}
			socketname = strdup(optarg);
			break;
		case 'r':
			hfo.scanroot = strdup(optarg);
			hfo.scanrootlen = strlen(hfo.scanroot);
			break;
		case 'd':{
				int len;
				len = strlen(optarg);
				if (optarg[len - 1] == '/') {
					quarantainedir = strdup(optarg);
				} else {
					quarantainedir = malloc(len + 2);
					memset(quarantainedir, 0, (len + 2));
					quarantainedir = strncpy(quarantainedir, optarg, len);
					quarantainedir[len] = '/';
				}
			}
			break;
		case 'i':
			pidfile = strdup(optarg);
			break;
		case 't':
			maxthreads = atoi(optarg);
			break;
		case 'g':
			maxlargethreads = atoi(optarg);
			break;
		case 'q':
			hfo.queue_limit = atoi(optarg);
			break;
		case 'm':
			status_timer = atoi(optarg);
			break;
		case 'l':
			loglevel = atoi(optarg);
			break;
		case 'b':
			bigfiles = atoi(optarg);
			break;
		case 'e':{
				int retval;
				hfo.excludepattern = strdup(optarg);
				retval = regcomp(&hfo.reg_pat, hfo.excludepattern, REG_EXTENDED | REG_NOSUB);
				if (retval == 0) {
					/* correct */
				} else {
					syslog(LOG_CRIT,
						   "Warning: exclude pattern %s is not a valid regular expression pattern!",
						   hfo.excludepattern);
					free(hfo.excludepattern);
					hfo.excludepattern = NULL;
				}
			}
			break;
		case 'n':
			detach_from_terminal = 0;
			break;
		case 'h':
		case '?':
			print_usage();
			exit(3);
		}
	}

	if (domain_socket) {
		s = open("/proc/sys/net/unix/max_dgram_qlen", O_WRONLY);
		if (s > 0) {
			write(s, "200", 3);
			close(s);
		}
	}
	if (detach_from_terminal) {
		detach(pidfile);
	} else {
		write_pidfile(pidfile, getpid());
	}

	openlog("scannedonlyd_clamav", LOG_PID, LOG_DAEMON);

#ifndef DEBUG
	signal(SIGSEGV, sighandler);
	signal(SIGTERM, sighandler);
#endif

	if (loglevel == 0)
		setlogmask(LOG_MASK(LOG_EMERG) | LOG_MASK(LOG_ALERT) |
				   LOG_MASK(LOG_CRIT) | LOG_MASK(LOG_ERR) | LOG_MASK(LOG_WARNING));
	else if (loglevel == 1)
		setlogmask(LOG_MASK(LOG_EMERG) | LOG_MASK(LOG_ALERT) |
				   LOG_MASK(LOG_CRIT) | LOG_MASK(LOG_ERR) | LOG_MASK(LOG_WARNING) | LOG_MASK(LOG_NOTICE));
	else if (loglevel == 2)
		setlogmask(LOG_MASK(LOG_EMERG) | LOG_MASK(LOG_ALERT) |
				   LOG_MASK(LOG_CRIT) | LOG_MASK(LOG_ERR) |
				   LOG_MASK(LOG_WARNING) | LOG_MASK(LOG_NOTICE) | LOG_MASK(LOG_INFO));
	else if (loglevel == 3)
		setlogmask(LOG_MASK(LOG_EMERG) | LOG_MASK(LOG_ALERT) |
				   LOG_MASK(LOG_CRIT) | LOG_MASK(LOG_ERR) |
				   LOG_MASK(LOG_WARNING) | LOG_MASK(LOG_NOTICE) | LOG_MASK(LOG_INFO) | LOG_MASK(LOG_DEBUG));
	else
		setlogmask(0);			/* 0 means that the current logmask is not modified */
	/* ready for normal work */ 
	if (domain_socket) {
		struct sockaddr_un saun;
		if (socketname == NULL) {
			socketname = strdup("/var/lib/scannedonly/scan");
		}
		syslog(LOG_NOTICE, "Starting with socket %s and %d threads", socketname, maxthreads);
			/*
			 * Get a socket to work with.  This socket will
			 * be in the UNIX domain, and will be a
			 * datagram socket.
			 !!!!!!!!!!!! set /proc/sys/net/unix/max_dgram_qlen to 200 !!!!!!!!!!
			 */ 
		if ((s = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0) {
			syslog(LOG_CRIT, "ABORT, could not create socket: %s", strerror(errno));
			exit(1);
		}
		memset(&saun, 0, sizeof(saun));
		saun.sun_family = AF_UNIX;
		strcpy(saun.sun_path, socketname);
		unlink(socketname);
		if (bind(s, (struct sockaddr *) &saun, SUN_LEN(&saun)) < 0) {
			syslog(LOG_CRIT, "ABORT, could not bind socket to %s: %s", socketname, strerror(errno));
			exit(1);
		}
		chmod(socketname, 0666);
	} else {
			/* UDP/IP socket
			   increase sys.net.core.netdev_max_backlog ??
			   net.core.rmem_max 
			 */ 
		struct sockaddr_in sain;
		syslog(LOG_NOTICE, "Starting with udp port %d and %d threads", portnum, maxthreads);
		if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			syslog(LOG_CRIT, "ABORT, could not create socket: %s", strerror(errno));
			exit(1);
		}
		memset(&sain, 0, sizeof(sain));
		sain.sin_family = AF_INET;
		sain.sin_port = htons(portnum);
		sain.sin_addr.s_addr = INADDR_ANY;
		if (bind(s, (struct sockaddr *) &sain, sizeof(struct sockaddr_in)) < 0) {
			syslog(LOG_CRIT, "ABORT, could not bind socket to port %d: %s", portnum, strerror(errno));
			exit(1);
		}
	}
	if (hfo.scanroot)
		syslog(LOG_INFO, "Scanning all requests in %s", hfo.scanroot);
	else
		syslog(LOG_INFO,
			   "Scanning any requests anywhere on the filesystem, use --scanroot to restrict scanning!");
	syslog(LOG_INFO, "Files larger than %d Mb are scanned in the large file queue", bigfiles);
	if (hfo.excludepattern)
		syslog(LOG_INFO, "Using exclude pattern %s", hfo.excludepattern);
	{
		int sndsize = 524288, newsize;
		unsigned int size = sizeof(int);
		/* default size is 131072 */ 
			/*
			   sndsize = 262144; */ 
			if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *) &sndsize, (int) sizeof(sndsize)) != 0) {
			syslog(LOG_NOTICE,
					"Failed to increase socket buffer size to %d bytes: %s", sndsize, strerror(errno));
		} else {
			getsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *) &newsize, &size);
			if (newsize < sndsize) {
				syslog(LOG_NOTICE,
					   "Failed to increase socket buffer size. Requested %d bytes, got %d bytes\n",
					   sndsize, newsize);
			}
		}
	}
	
	/* load all available databases from default directory */
	syslog(LOG_INFO, "Now loading database...");
#ifdef HAVE_LIBCLAMAV_CL_INIT
	if (cl_init(CL_INIT_DEFAULT) != CL_SUCCESS) {
		syslog(LOG_CRIT, "ABORT, failed to initialize libclamav");
		exit(3);
	}
#endif
	engine[0] = load_engine();
	engine[1] = NULL;
	current_engine = 0;
	if (engine[0] == NULL) {
		syslog(LOG_CRIT, "ABORT, exiting: no clam engine");
		exit(2);
	}
	memset(&dbstat, 0, sizeof(struct cl_stat));
	cl_statinidir(cl_retdbdir(), &dbstat);
#ifndef HAVE_LIBCLAMAV_CL_INIT
	memset(&limits, 0, sizeof(struct cl_limits));
	limits.maxfiles = 1000;	/* max files within an archive */
	limits.maxfilesize = 50 * 1048576;	/* maximum size of archived/compressed
										 * file (files exceeding this limit
										 * will be ignored)
										 */
	limits.maxreclevel = 5;	/* maximum recursion level for archives */
	
#ifdef HAVE_LIBCLAMAV_LIMITS_MAXRATIO
	limits.maxmailrec = 64;	/* maximum recursion level for mail files */
	limits.maxratio = 200;		/* maximum compression ratio */
	
#else							/* HAVE_LIBCLAMAV_LIMITS_MAXRATIO */
	limits.maxscansize = 200 * 1048576;	/* during the scanning of archives this
											 * size will never be exceeded
											 */
#endif							/* HAVE_LIBCLAMAV_LIMITS_MAXRATIO */
#endif							/* HAVE_LIBCLAMAV_CL_INIT */
	prev_status_update.tv_sec = time(NULL);
	prev_status_update.tv_nsec = 0;
	next_status_update.tv_sec = prev_status_update.tv_sec + 60 * status_timer;
	next_status_update.tv_nsec = 0;
	next_dbcheck.tv_sec = prev_status_update.tv_sec + 60 * dbcheck_timer;
	next_dbcheck.tv_nsec = 0;
	small_queue = new_queue();
	large_queue = new_queue();
	pthread_mutex_init(&queuelock, NULL);
	pthread_mutex_init(&enginelock, NULL);
	pthread_spin_init(&statuslock, 0);
	currently_scanning = malloc((maxthreads) * sizeof(char *));
	memset(currently_scanning, 0, (maxthreads) * sizeof(char *));
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	for (i = 0; i < maxthreads; i++) {
		pthread_create(&thread, &attr, (void *) &scanningthread, INT_TO_POINTER(i));
	}
	syslog(LOG_NOTICE, "Finished building databases. Ready for scanning with %d threads", maxthreads);
	for (;;) {
		char *cur, *next;
		len = recv(s, buffer, BUFFERSIZE - 1, 0);	/* while we endless loop we block on this call listening for requests */
		if (len == -1) {
			syslog(LOG_NOTICE, "Error reading from socket: %s", strerror(errno));
			continue;
		}
		buffer[len] = '\0';
		if (len <= 1) {
			syslog(LOG_INFO, "Ignored very short request '%s'; not a filename", buffer);
			continue;
		}
		DEBUG_MSG("received buffer %s\n", buffer);
		cur = buffer;
		next = strchr(buffer, '\n');
		do {
			if (next) {
				*next = '\0';
			}

			DEBUG_MSG("extracted filename %s\n", cur);
			handle_filename(&hfo, cur);

			if (next && next + 1 - buffer < len) {
				cur = next + 1;
				if (cur[0] == '\0') {
					cur = NULL;
				} else {
					next = strchr(cur, '\n');
				}
			} else {
				cur = NULL;
			}
		} while (cur);
	}
	pthread_attr_destroy(&attr);
#ifdef HAVE_LIBCLAMAV_CL_INIT
	if (engine[0])
		cl_engine_free(engine[0]);
	if (engine[1])
		cl_engine_free(engine[1]);
#else	/*  */
	if (engine[0])
		cl_free(engine[0]);
	if (engine[1])
		cl_free(engine[1]);
#endif		/*  */
	free(small_queue);
	free(large_queue);
}
